from collections import namedtuple


## catches, est_zippin, est_carle_strub, cathability: q = 1-p,
Catch = namedtuple("Catch", "data hat_nz hat_ncs p x ci")

# note: x is the _weighted_sum defined by
#
# lambda data : sum((len(data) - (1 + i)) * e for (i, e) in enumerate(data))

catches = []

#
# 0
#
catches.append(
    Catch(
        data=[100, 10, 1, 1, 0, 0],
        hat_nz=112,
        hat_ncs=112,
        p=0.9,
        x=-1,
        ci=1,
    )  ## ???
)


#
# 1
#
catches.append(
    Catch(
        data=[19, 17, 13, 1, 1],
        hat_nz=53,
        hat_ncs=53,
        p=0.46,
        x=154,
        ci=4.4,
    )
)


#
# 2
#
catches.append(
    Catch(
        data=[10, 20, 30, 24, 2, 7],
        hat_nz=156,
        hat_ncs=145,
        p=0.14,
        x=270,
        ci=80.29,
    )
)

#
# 3
#
catches.append(
    Catch(
        data=[5, 1, 0, 0, 0, 0, 0],
        hat_nz=6,
        hat_ncs=6,
        p=0.86,
        x=35,
        ci=0.00529,
    )
)


#
# 4
#
catches.append(
    Catch(
        data=[34, 46, 22, 26, 18, 16, 20, 12],
        hat_nz=268,
        hat_ncs=264,
        p=0.1454,
        x=834,
        ci=56.33,
    )
)

# laks (5)
catches.append(
    Catch(
        data=[32, 40, 12, 19, 9, 7, 8, 5, 2, 3, 1, 1, 0],
        hat_nz=141,
        hat_ncs=140,
        p=None,
        x=None,
        ci=3.38,
    )
)

# orre (6)
catches.append(
    Catch(
        data=[22, 9, 3, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0],
        hat_ncs=38,
        hat_nz=38,
        p=None,
        x=None,
        ci=0.06,
    )
)



def _weighted_sum(data):
    return sum((len(data) - (1 + i)) * e for (i, e) in enumerate(data))


def _z_pre_estimate(data, hatN):
    t = sum(data)
    x = _weighted_sum(data)
    k = len(data)

    tellerA = hatN - t + 0.5
    tellerB = (k * hatN - x) ** k
    nevner = (k * hatN - x - t) ** k

    return (tellerA * tellerB) / nevner - 0.5


def _cs_Eq(t, hatN, k, x):
    prod = 1
    for i in range(k):
        j = i + 1
        teller = k * hatN - x - t + 1 + k - j
        nevner = k * hatN - x + 2 + k - j
        prod *= (teller * 1.0) / nevner

    return t - 1 + ((hatN + 1) * prod)


def removal_carle_strub(data):
    t = sum(data)
    x = _weighted_sum(data)
    k = len(data)
    hatN = t

    for i in range(1000 * 1000):
        lhs = hatN + i
        rhs = _cs_Eq(t, lhs, k, x)
        if lhs >= rhs:
            return lhs
    raise ValueError("Unable to find CS solution")


assert removal_carle_strub([34, 46, 22, 26, 18, 16, 20, 12]) == 264


def removal_zippin(data):
    t = sum(data)
    k = len(data)
    x = _weighted_sum(data)
    z_min = ((t - 1) * (k - 1) / 2) - 1

    if x <= z_min:
        raise ValueError(f"Zippin X below z_min for {data}")

    hatN = t
    for i in range(1000 * 1000):
        lhs = hatN + i
        rhs = _z_pre_estimate(data, lhs)
        if rhs > lhs:
            return lhs

    raise ValueError("Unable to find CS solution")


assert removal_zippin([34, 46, 22, 26, 18, 16, 20, 12]) == 268


def main():
    from sys import argv

    if len(argv) < 2:
        exit("Usage: carle_strub 30 20 10 0")

    data = [int(e.lstrip(",").rstrip(",")) for e in argv[1:]]
    print("cs:", removal_carle_strub(data))
    print("z: ", removal_zippin(data))


if __name__ == "__main__":
    main()



"""Removal sampling.

Maximum Likelihood (ML) estimates for estimating population size (N)
for a constant probability of capture model.
"""

import sys
import statistics
import random

import numpy as np
import pymc3 as pm
import matplotlib.pyplot as plt


def exit_with_usage():
    msg = """Usage:
removal_sampling.py --plot --testcase 3
removal_sampling.py --data 25 10 2 3 0 1 0 0
removal_sampling.py --samples 1000 --data 25 10 2 3 0 1 0 0
removal_sampling.py --tuning 5000 --data 25 10 2 3 0 1 0 0
removal_sampling.py --savefig --data 25 10 2 3 0 1 0 0
removal_sampling.py --beta --data 25 10 2 3 0 1 0 0
"""
    exit(msg)


def _savefig(pfx=""):
    fnameid = random.randint(10 ** 12, 10 ** 15)
    fname = f"{pfx}{fnameid}.svg"
    plt.savefig(fname)
    plt.clf()
    return fname


def _get_testcase(n):
    from .testdata import catches

    return catches[n].data


def run(samples=None, tune=None, data=None, testcase=None):
    if samples is None:
        samples = 5000
    if tune is None:
        tune = 1000
    if data is None and testcase is None:
        raise ValueError("Specify data or testcase")
    if testcase is not None:
        case = _get_testcase(testcase)
    else:
        case = list(data)

    print(f"data    = {case}")
    print(f"samples = {samples}")
    print(f"tunings = {tune}")

    rm_model = pm.Model()

    with rm_model:

        N = pm.DiscreteUniform("N", lower=0, upper=500)
        p = pm.Uniform("p", lower=0, upper=1)

        observations = case
        catch = []
        for idx, c in enumerate(observations):
            q = pm.Binomial(f"q{idx}", N - sum(catch), p, observed=[c])
            catch.append(q)
        trace = pm.sample(draws=samples, tune=tune)
    return rm_model, trace


def summary(model, trace):
    with model:
        sorted_N = sorted(trace["N"])
        len_N = len(trace["N"])

        quantiles = statistics.quantiles(sorted_N)
        quantiles_s = [25, 50, 75]
        print(
            "   ".join(
                [f"{i}% = {round(q, 1)}" for q, i in zip(quantiles, quantiles_s)]
            )
        )

        return pm.summary(trace), sorted_N


def _plot(model, trace, savefig=False):
    with model:

        pm.traceplot(trace, ["N", "p"])
        if savefig:
            return _savefig("smry")
        else:
            plt.show()


def _get_arg(args, arg):
    if arg == "--data" and arg in args:
        idx = args.index(arg)
        return [int(x.rstrip(",").lstrip(",")) for x in args[idx + 1 :]]
    if arg in args:
        return int(args[args.index(arg) + 1])
    return None


def _beta(data, plot=False, savefig=False):
    try:
        from scipy.stats import beta
    except ImportError:
        exit("--beta needs scipy")

    tens = len(data) // 50
    data = data[tens:-tens]

    alpha_, beta_, loc_, scale_ = [round(x, 3) for x in beta.fit(data)]
    print(f"X = Beta(alpha={alpha_}, beta={beta_}, loc={loc_}, scale={scale_})")
    print(f"E[X] = a/(a+b) = {alpha_/(alpha_+beta_)+loc_}")

    if not (plot or savefig):
        return alpha_, beta_, loc_, scale_

    x = np.linspace(
        beta.ppf(0.01, alpha_, beta_, loc_, scale_),
        beta.ppf(0.99, alpha_, beta_, loc_, scale_),
        100,
    )

    plt.plot(
        x,
        beta.pdf(x, alpha_, beta_, loc=loc_, scale=scale_),
        label=f"Beta({alpha_}, {beta_})",
    )
    plt.hist(
        data, alpha=0.75, color="green", bins=min(200, len(set(data))), density=True
    )
    if savefig:
        return (alpha_, beta_, loc_, scale_), _savefig("beta")
    plt.show()
    return alpha_, beta_, loc_, scale_


def main():
    args = sys.argv
    if len(args) == 1:
        exit_with_usage()

    if "--testcase" in args and "--data" in args:
        exit_with_usage()

    testcase = _get_arg(args, "--testcase")
    data = _get_arg(args, "--data")
    samples = _get_arg(args, "--samples")
    tune = _get_arg(args, "--tune")

    model, trace = run(testcase=testcase, samples=samples, tune=tune, data=data)
    smry, sorted_N = summary(model, trace)
    print(smry)
    if "--plot" in args:
        _plot(model, trace)
    elif "--savefig" in args:
        fname = _plot(model, trace, savefig=True)
        print(f"wrote {fname}")

    if "--beta" in args:
        retval = _beta(sorted_N, plot="--plot" in args, savefig="--savefig" in args)
        if len(retval) == 2:
            fname = retval[1]
            retval = retval[0]
            print(f"wrote {fname}")
        print(f"Beta({retval})")


if __name__ == "__main__":
    main()
